/******************************************************************************
* HTTP Dmon                                                                  *
* Projekt Informatik Klasse 11                                                *
*                                                                             *
* Anleitung:                                                                  *
*   - Kompilieren                                                             *
*   - Ausfhren                                                               *
*   - Browser ffnen auf URL                                                  *
*     http://localhost:<port>/                                                *
*     --> <port> durch den angezeigten                                        *
*         Port ersetzen                                                       *
*   - Dokumentation folgen                                                    *
*                                                                             *
*                                                                             *
* Copyright (c) 2005 by                                                       *
* Phillip Berndt                                                              *
*                                                                             *
******************************************************************************/

// Bentigte Bibliotheken einbinden
import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.*;

// Dies ist ein Konsolenprogramm, kein Applet
// Applets wren aufgrund der eingeschrnkten Dateizugriffsrechte
// ungeeignet
public class JavaHTTPd
{
  // Einstellungen werden statisch auf Klassenebene gespeichert
  static String setting_RootDirectory;
  static int setting_ConnectionTimeout;
  static HashMap setting_DirectoryIndexes;
  static HashMap setting_CGIHandler;
  static int setting_serverPort;
  static long setting_maxPostLength;

  // Dies ist die Hauptprozedur dieses Programms
  // Sie kmmert sich um eingehende Verbindungen
  public static void main(String args[]) throws IOException
  {
    JavaHTTPd.out("JavaHTTPd");
    JavaHTTPd.out("   Copyright (c) 2005 by Phillip Berndt");

    // Variablen mit Defaultwerten initialisieren
    JavaHTTPd.setting_DirectoryIndexes = new HashMap();
    JavaHTTPd.setting_CGIHandler = new HashMap();
    JavaHTTPd.setting_ConnectionTimeout = 10;
    JavaHTTPd.setting_RootDirectory = "wwwroot";
    JavaHTTPd.setting_serverPort = 80;

    // Konfigurationsdatei laden
    // Dafr eine Fehlerbehandlung bieten
    try
    {
      if(!(new File("JavaHTTPd.config")).isFile())
      {
        JavaHTTPd.out("JavaHTTPd.config existiert nicht");
        return;
      }
      setting_RootDirectory = "";
      InputStreamReader stream_configFile = (new InputStreamReader((new FileInputStream("JavaHTTPd.config"))));
      char[] buffer_configFile = new char[ (int) (new File("JavaHTTPd.config")).length()];
      stream_configFile.read(buffer_configFile);
      String string_configFile = String.copyValueOf(buffer_configFile);
      stream_configFile.close();
      StringTokenizer token_configFile = new StringTokenizer(string_configFile, "\n");

      // Konfigurationsdatei einlesen und verarbeiten
      while (token_configFile.hasMoreElements())
      {
        String configLine = token_configFile.nextToken().trim();
        if(configLine.indexOf(' ') < 1) continue;
        String command = configLine.substring(0, configLine.indexOf(' '));

        // Basisverzeichnis des Webservers
        if (command.compareTo("RootDirectory") == 0)
          JavaHTTPd.setting_RootDirectory = configLine.substring(14);

        // Verbindungstimeout
        if (command.compareTo("ConnectionTimeout") == 0)
           JavaHTTPd.setting_ConnectionTimeout = Integer.parseInt(configLine.substring(18));

        // Standardport des Servers; Fallback mglich
        if (command.compareTo("ServerPort") == 0)
           JavaHTTPd.setting_serverPort = Integer.parseInt(configLine.substring(11));

        // Maximale Lnge fr Postdaten
        if (command.compareTo("MaxPostlength") == 0)
            JavaHTTPd.setting_maxPostLength = Math.abs(Integer.parseInt(configLine.substring(14))) * 1024 * 1024;

        // Indexdateien fr Verzeichnisse
        if (command.compareTo("DirectoryIndex") == 0)
        {
          StringTokenizer dirIndex = new StringTokenizer(configLine.substring(15), " ");
          while (dirIndex.hasMoreElements())
            JavaHTTPd.setting_DirectoryIndexes.put(dirIndex.nextToken(), "-");
        }

        // CGI (Common gateway interface) Programme
        if (command.compareTo("AddHandler") == 0)
        {
          String handlerLine = configLine.substring(11);
          String regex = handlerLine.substring(0, handlerLine.indexOf(" "));
          String cmd = handlerLine.substring(handlerLine.indexOf(" ") + 1);
          JavaHTTPd.setting_CGIHandler.put(regex, cmd);
        }
      }
    }
    catch(Exception fileParseException)
    {
      // Bei Fehlern abbrechen
      JavaHTTPd.out("Syntax der Konfigurationsdatei ngltig");
      return;
    }

    // Prfen, ob htdocs existiert
    if(!(new File(JavaHTTPd.setting_RootDirectory)).isDirectory())
    {
      JavaHTTPd.out(JavaHTTPd.setting_RootDirectory + " existiert nicht");
      return;
    }
    JavaHTTPd.setting_RootDirectory = (new File(JavaHTTPd.setting_RootDirectory)).getAbsolutePath() + "/";

    // Einen Server erstellen.
    // Standard fr HTTP ist Port 80/TCP, dennoch
    // einen Fallback bereithalten
    ServerSocket httpServer = new ServerSocket();
    JavaHTTPd.out("Erstelle Socket auf Port " + JavaHTTPd.setting_serverPort);
    boolean bindingSucceeded;
    do
    {
      bindingSucceeded = true;
      try
      {
        // Hier wird der TCP Server erstellt
        httpServer = new ServerSocket(JavaHTTPd.setting_serverPort);
      }
      catch(IOException httpErrorException)
      {
        // Erneut versuchen
        JavaHTTPd.setting_serverPort++;
        JavaHTTPd.out("WARNUNG: Fallback auf Port " + JavaHTTPd.setting_serverPort);
        bindingSucceeded = false;
      }
    }
    while(!bindingSucceeded);


    // Nachdem nun feststeht, auf welchem Port der Server wartet, ist er
    // im "Wartemodus" - nun kann er eingehende Verbindungen akzeptieren
    // Damit er jedoch mit mehreren Anfragen gleichzeitig umgehen kann
    // ist Multithreading gefragt.
    while(true)
    {
      // Accept wartet auf eine Verbindung und speichert sie in einer Variable
      Socket incommingConnection = httpServer.accept();

      // Um die Verbindung abzuhandeln - wie bereits gesagt - benutze ich Multithreading
      HTTPServerSocket handler = new HTTPServerSocket(incommingConnection);
      Thread socketHandler = new Thread(handler);
      // setDaemon Macht einen Dmonen aus dem Thread
      // Das sorgt dafr, dass er im Hintergrund abluft
      socketHandler.setDaemon(true);
      socketHandler.start();
    }
  }

  // Hier folgen Prozeduren, die statisch genutzt werden
  // Sie betreffen vorallem Statusausgabe und bestimmte Routinen,
  // die hufig bentigt werden

  // Eine Zeile ausgeben
  // Mit aktueller Zeit
  public static void out(String line)
  {
    Calendar calendar = Calendar.getInstance();

    // Erst einmal die Zeit ausgeben
    // Etwas lnger, da am Anfang evntl. Nullen angefgt werden mssen
    System.out.print(
      "[" + (calendar.get(Calendar.HOUR_OF_DAY) >= 10 ? "" : "0") +
       calendar.get(Calendar.HOUR_OF_DAY) + ":" +
       (calendar.get(Calendar.MINUTE) >= 10 ? "" : "0") +
       calendar.get(Calendar.MINUTE) +
      ":" + (calendar.get(Calendar.SECOND) >= 10 ? "" : "0") +
      calendar.get(Calendar.SECOND) + "] ");

    // Nun die Statusmeldung
    System.out.println(line);
  }

  // Eine Datei aus dem htdocs Ordner laden
  // Dabei nicht als String lesen, da Java dann den Charset ndert!!
  public static byte[] getFile(String fileName)
  {
    byte[] returnedFile = new byte[0];

    try
    {
        FileInputStream data = new FileInputStream(JavaHTTPd.setting_RootDirectory + fileName);
        returnedFile = new byte[data.available()];
        data.read(returnedFile);
        data.close();
        return returnedFile;
    }
    catch(IOException exception)
    {
        return returnedFile;
    }
  }

  // Einen Text HTML verschlsseln
  public static String htmlSpecialChars(String text)
  {
    return text.replaceAll("<", "&lt;").replaceAll(">", "&gt;").replaceAll("\"", "&quot;");
  }

  // Aus einem Byte Array einen String extrahieren
  public static String bufferExtract(byte[] buffer, int bytesRead)
  {
    String retBuffer = "";
    for(int currByte=0; currByte<bytesRead; currByte++)
      retBuffer += (char)buffer[currByte];
    return retBuffer;
  }
}

// Diese Klasse kmmert sich um die Verbindungen
class HTTPServerSocket implements Runnable
{
  Socket requestSocket;
  InputStream HTTPSocketRead;
  OutputStream HTTPSocketWrite;

  // Konstruktor des Threads
  public HTTPServerSocket(Socket requestSocket) throws IOException
  {
    // Dem Socket Variablen zuweisen
    this.requestSocket = requestSocket;
    // Ich wrde gerne Streamwrapper benutzen - leider verflschen diese
    // das Charset
    this.HTTPSocketRead = requestSocket.getInputStream();
    this.HTTPSocketWrite = requestSocket.getOutputStream();
  }

  // Destruktor - schliet die Verbindung
  public void finalize()  throws IOException
  {
    // Sicherstellen, dass alles gesendet wurde
    HTTPSocketWrite.flush();
    this.requestSocket.close();
  }

  // Wird ausgefhrt, wenn der Thread gestartet wird
  public void run()
  {
    // Da wir es hier mit potentiellen Hackern zu tun haben knnen
    // zumindest eine Fehlerbehandlung implementieren
    try
    {
      // Ein HTTP Request endet mit 2x CRLF
      // Bis dahin lesen, jedoch ein Timeout festlegen
      String HTTPRequest = "";
      long timeOut = Calendar.getInstance().getTime().getTime() + JavaHTTPd.setting_ConnectionTimeout * 1000;
      while(true)
      {
        // Bytes lesen
        if (this.HTTPSocketRead.available() != 0)
          HTTPRequest += (char)this.HTTPSocketRead.read();
        // Wenn CRLF CRLF enthalten ist, ist die Anfrage vollstndig
        if (HTTPRequest.indexOf("\r\n\r\n") >= 0) break;
        // Auf Timeout prfen
        if (Calendar.getInstance().getTime().getTime() > timeOut)
        {
          // Fehler 408 = Timeout
          this.errorDie(408);
          return;
        }
        // Die Kontrolle abgeben, um Prozessorleistung zu sparen
        Thread.yield();
      }

      // Zu diesem Zeitpunkt ist das Header der Anfrage vollstndig
      // Nun muss sie komplett verarbeitet werden

      // Die Anfrage unterteilen
      StringTokenizer HTTPRequestData = new StringTokenizer(HTTPRequest, "\r\n");

      // Die 1. Zeile ist immer das HTTP Request
      String HTTPTask = HTTPRequestData.nextToken();

      // String.matches nutzt sogenannte Regular Expressions
      // .+ steht hier fr: Ein beliebiges Zeichen (.), beliebig oft wiederholt (+)
      // (1|0) steht fr: Entweder 1 oder 0, [...]
      // Mgliche Typen fr diese Anfrage sind:
      // GET -> Eine Datei anfordern
      // POST -> Eine Datei anfordern, bei der Anfrage jedoch Daten senden
      // HEAD -> Wie GET, jedoch nur den Header ermitteln
      // HTTP/1.1 implementiert noch weitere Typen, hier sind jedoch nur diese
      // Grundtypen integriert.
      if(!HTTPTask.matches("^(GET|POST|HEAD) .+ HTTP/1.(1|0)"))
      {
        // Fehler 405 = Ungltiges Header
        this.errorDie(405);
        return;
      }

      // HEAD bedeutet: Nur den Header ausgeben. Dies vermerken...
      boolean headerOnly = HTTPTask.matches("^HEAD.+");

      // Die Datei extrahieren
      // Dabei URL Kodierung aufheben
      String HTTPTaskFile = HTTPTask.substring(HTTPTask.indexOf(" ") + 1, HTTPTask.indexOf(" ", 5));
      HTTPTaskFile = URLDecoder.decode(HTTPTaskFile, "ISO-8859-1");

      // Wenn ein Query existiert (Hinter einem ?), dieses abschneiden
      String queryString = "";
      if(HTTPTaskFile.indexOf('?') > 0)
      {
        queryString = HTTPTaskFile.substring(HTTPTaskFile.indexOf('?') + 1);
        HTTPTaskFile = HTTPTaskFile.substring(0, HTTPTaskFile.indexOf('?'));
      }

      // Jedliche ../ entfernen --> Sicherheitslcke !!
      // Hier werden wieder Regular Expressions benutzt
      HTTPTaskFile = HTTPTaskFile.replaceAll("\\.\\.(/|\\\\)", "");

      // Darauf folgen Zeilen nach dem Schema
      // ATTRIBUT: WERT
      // Die so genannten HTTP Header
      HashMap HTTPRequestHeaders = new HashMap(3);
      while(HTTPRequestData.hasMoreElements())
      {
        String[] HTTPNext = HTTPRequestData.nextToken().split(": ", 2);
        if(HTTPNext.length != 2)
        {
          // Fehler 400 = Ungltige Anfrage
          this.errorDie(400);
          return;
        }
        HTTPRequestHeaders.put(HTTPNext[0].toLowerCase(), HTTPNext[1]);
      }

      // Da der Host Header Pflicht ist, prfen, ob er gesetzt ist
      // Ihn aber ansonsten ignorieren, da dies ein Ein-Host-Server ist
      if(!HTTPRequestHeaders.containsKey("host"))
      {
        // Fehler 400 = Ungltige Anfrage
        this.errorDie(400);
        return;
      }

      // Wenn POST genutzt wurde, alle Daten aus dem Body laden
      byte[] postData = new byte[0];
      if(HTTPTask.matches("^POST.+") == true && HTTPRequestHeaders.containsKey("content-length"))
      {
        // Lnge der Postdaten ermitteln
        int dataLength = Integer.parseInt((String)HTTPRequestHeaders.get("content-length"));
        // Lngenbeschrnkung beachten
        if(dataLength < 0 || dataLength > JavaHTTPd.setting_maxPostLength)
        {
          // Fehler 400 = Ungltige Anfrage
          this.errorDie(400);
          return;
        }
        postData = new byte[dataLength];
        // Wieder ein Timeout definieren
        timeOut = Calendar.getInstance().getTime().getTime() + JavaHTTPd.setting_ConnectionTimeout * 1000;
        int postDataBytesRead = 0;
        while(true)
        {
          // Bytes lesen
          if(this.HTTPSocketRead.available() > 0)
          {
            postDataBytesRead += this.HTTPSocketRead.read(postData, postDataBytesRead, dataLength - postDataBytesRead);
            timeOut = Calendar.getInstance().getTime().getTime() + 10000;
          }

          if(this.HTTPSocketRead.available() < 0) return; // Abbruch?

          // Wenn postData die geforderte Lnge hat abbrechen
          if(postDataBytesRead >= dataLength)
            break;

          // Auf Timeout prfen
          if (Calendar.getInstance().getTime().getTime() > timeOut)
          {
            // Fehler 408 = Timeout
            this.errorDie(408);
            return;
          }

          // Die Kontrolle abgeben, um Prozessorleistung zu sparen
          Thread.yield();
        }
      }

      // An diesem Punkt ist alles wichtige bekannt:
      // Der Dateiname und die Header
      //
      File requestFile = new File(JavaHTTPd.setting_RootDirectory + "htdocs/" + HTTPTaskFile);
      byte[] fileData;

      // Bei Verzeichnissen sollte das "/" nicht fehlen!!
      if(requestFile.isDirectory() && (HTTPTaskFile.substring(HTTPTaskFile.length() - 1).compareTo("/") != 0))
      {
        // Statuscode 301 steht fr Weiterleitungen
        HTTPSocketWrite.write(this.genHTTPHeader(301, 0).getBytes());
        String redirection = "Location: " + HTTPTaskFile + "/\r\n\r\n" +
            "<a href=\"" + HTTPTaskFile + "/\">Click</a>";
        HTTPSocketWrite.write(redirection.getBytes());
        return;
      }

      // Bei Verzeichnissen:
      if(requestFile.isDirectory())
      {
        // Existiert eine Indexdatei?
        Object[] indexFiles = JavaHTTPd.setting_DirectoryIndexes.keySet().toArray();
        for(int indexFilesi=0; indexFilesi<indexFiles.length; indexFilesi++)
        {
          String defaultFile = (String)indexFiles[indexFilesi];
          if((new File(JavaHTTPd.setting_RootDirectory + "htdocs/" + HTTPTaskFile + "/" + defaultFile)).exists())
          {
            // Wenn ja - dorthin weiterleiten
            HTTPTaskFile += "/" + defaultFile;
            requestFile = new File(JavaHTTPd.setting_RootDirectory + "htdocs/" + HTTPTaskFile);
            break;
          }
        }
      }

      // Dateityp prfen
      if(requestFile.isDirectory())
      {
        // Es handelt sich um ein Verzeichnis
        // -> Verzeichnisse werden einfach als Linkliste ausgebeben
        //    Der HTML Quellcode dazu steht hier fest verankert, da er sehr
        //    kurz ist
        if(HTTPTaskFile.charAt(HTTPTaskFile.length() - 1) != '/')
          HTTPTaskFile += "/";

        // HTML Header
        String outputIndex = "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">"+
            "<html><head><style>*{font-family:monospace}</style><title>Directory Index for " +
            JavaHTTPd.htmlSpecialChars(HTTPTaskFile) + "</title></head><body>" +
            "<h1>Directory Index for " + JavaHTTPd.htmlSpecialChars(HTTPTaskFile) +
            "</h1><ul><li>[DIRECTORY] <a href=\"../\">..</a></li>";

        // Dateien erst UNTER Verzeichnissen ausgeben
        String outputIndex_Files = "";

        // Alle Dateien durchlaufen
        for(int index=0; index < requestFile.list().length; index++)
        {
          if((new File(JavaHTTPd.setting_RootDirectory + "htdocs/" + HTTPTaskFile + "/" + requestFile.list()[index])).isDirectory())
          {
            // Verzeichnis ausgeben
            outputIndex += "<li>[DIRECTORY] <a href=\"" +
                JavaHTTPd.htmlSpecialChars(HTTPTaskFile) +
                JavaHTTPd.htmlSpecialChars(requestFile.list()[index]) + "\">" +
                JavaHTTPd.htmlSpecialChars(requestFile.list()[index]) +
                "</a></li>";
          }
          else
          {
            // Datei ausgeben
            outputIndex_Files += "<li><a href=\"" +
                JavaHTTPd.htmlSpecialChars(HTTPTaskFile) +
                JavaHTTPd.htmlSpecialChars(requestFile.list()[index]) + "\">" +
                JavaHTTPd.htmlSpecialChars(requestFile.list()[index]) +
                "</a></li>";
          }
        }

        // Dateien anhngen
        outputIndex += outputIndex_Files;

        // HTML Footer
        outputIndex += "</ul><hr><address>JavaHTTPd by Phillip Berndt</address></body></html>";

        // Jetzt den Header und die Daten ausgeben
        HTTPSocketWrite.write(this.genHTTPHeader(200, outputIndex.length()).getBytes());
        if(headerOnly)
        {
          HTTPSocketWrite.flush();
          HTTPSocketWrite.close();
          return;
        }
        HTTPSocketWrite.write(outputIndex.getBytes());
        HTTPSocketWrite.flush();
        HTTPSocketWrite.close();
      }
      else if(requestFile.isFile())
      {
        // Es handelt sich um eine Datei
        // CGI Scripte ausfhren
        Object[] cgiHandlers = JavaHTTPd.setting_CGIHandler.keySet().toArray();
        for(int cgiHandlersi=0; cgiHandlersi<cgiHandlers.length; cgiHandlersi++)
        {
          String cgiMatch = (String)cgiHandlers[cgiHandlersi];
          if(HTTPTaskFile.matches(cgiMatch))
          {
            String cgiHandler = (String)JavaHTTPd.setting_CGIHandler.get(cgiHandlers[cgiHandlersi]);
            JavaHTTPd.out("(" + this.requestSocket.getInetAddress().getHostName() + ") Script[" + cgiMatch + "]> " + HTTPTaskFile);
            this.useCGI(cgiHandler, HTTPTaskFile, queryString, HTTPRequestHeaders, postData, headerOnly);
            return;
          }
        }

        // Handelt es sich um eine normale Datei, Datei lesen
        JavaHTTPd.out("(" + this.requestSocket.getInetAddress().getHostName() + ") File> " + HTTPTaskFile);

        // Hier kein getFile benutzen, sondern die Daten direkt laden
        // Das spart Speicher, da in Blcken gelesen werden kann
        HTTPSocketWrite.write(this.genHTTPHeader(200, requestFile.length()).getBytes());

        // Nur den Header? Dann die Ausgabe nun beenden
        if(headerOnly)
        {
          HTTPSocketWrite.flush();
          HTTPSocketWrite.close();
          return;
        }

        // Ansonsten die Datei in 1kb Blcken lesen
        FileInputStream fileDataStream = new FileInputStream(JavaHTTPd.setting_RootDirectory + "htdocs/" + HTTPTaskFile);
        while(true)
        {
          byte buffer[] = new byte[1024];
          int currentCharCount = fileDataStream.read(buffer);
          if(currentCharCount < 1) break; // End of file
          HTTPSocketWrite.write(buffer, 0, currentCharCount);
        }
        HTTPSocketWrite.flush();
        HTTPSocketWrite.close();
        fileDataStream.close();
      }
      else
      {
         // Die Datei wurde nicht gefunden - 404
         this.errorDie(404);
         return;
      }
    }
    catch(Exception SocketException)
    {
      // Wenn mglich, Fehler 500 senden
      if(this.requestSocket.isConnected())
      {
        // Einen evntl. Fehler zwar abfangen, darauf reagieren ist aber nicht mglich/ntig
        try {  this.errorDie(500); }
        catch(Exception errException) { }
      }

      // Fehlermeldung
      JavaHTTPd.out("Fehler whrend Verbindung: " + SocketException);
    }
  }

  // Funktionen fr die HTTP Behandlung
  void errorDie(int errorNumber) throws IOException
  {
    // Fehlerdatei laden & mit Header ausgeben
    byte[] errorText = JavaHTTPd.getFile("error/" + errorNumber + ".html");
    HTTPSocketWrite.write(this.genHTTPHeader(errorNumber, (long)errorText.length).getBytes());
    HTTPSocketWrite.write(errorText);
    HTTPSocketWrite.flush();
    finalize();
  }

  // HTTP Antwort Header erzeugen
  // Leider ohne "Content-Type" Header; ich wei nicht, wie man mit
  // Java (ohne Hilfsmittel) den Mimetyp einer Datei ermittelt.
  String genHTTPHeader(int statusCode, long bodyLength)
  {
    String httpHeader = "";
    httpHeader += "HTTP/1.1 " + statusCode + " ";

    // Eine vollstndige Liste findet sich auf
    // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
    // Umgeformt mithilfe von jEdit, dem genialsten Editor berhaupt fr
    // Java Code
    switch(statusCode)
    {
      case 100: httpHeader += "Continue"; break;
      case 101: httpHeader += "Switching Protocols"; break;
      case 200: httpHeader += "OK"; break;
      case 201: httpHeader += "Created"; break;
      case 202: httpHeader += "Accepted"; break;
      case 203: httpHeader += "Non-Authoritative Information"; break;
      case 204: httpHeader += "No Content"; break;
      case 205: httpHeader += "Reset Content"; break;
      case 206: httpHeader += "Partial Content"; break;
      case 300: httpHeader += "Multiple Choices"; break;
      case 301: httpHeader += "Moved Permanently"; break;
      case 302: httpHeader += "Found"; break;
      case 303: httpHeader += "See Other"; break;
      case 304: httpHeader += "Not Modified"; break;
      case 305: httpHeader += "Use Proxy"; break;
      case 306: httpHeader += "(Unused)"; break;
      case 307: httpHeader += "Temporary Redirect"; break;
      case 400: httpHeader += "Bad Request"; break;
      case 401: httpHeader += "Unauthorized"; break;
      case 402: httpHeader += "Payment Required"; break;
      case 403: httpHeader += "Forbidden"; break;
      case 404: httpHeader += "Not Found"; break;
      case 405: httpHeader += "Method Not Allowed"; break;
      case 406: httpHeader += "Not Acceptable"; break;
      case 407: httpHeader += "Proxy Authentication Required"; break;
      case 408: httpHeader += "Request Timeout"; break;
      case 409: httpHeader += "Conflict"; break;
      case 410: httpHeader += "Gone"; break;
      case 411: httpHeader += "Length Required"; break;
      case 412: httpHeader += "Precondition Failed"; break;
      case 413: httpHeader += "Request Entity Too Large"; break;
      case 414: httpHeader += "Request-URI Too Long"; break;
      case 415: httpHeader += "Unsupported Media Type"; break;
      case 416: httpHeader += "Requested Range Not Satisfiable"; break;
      case 417: httpHeader += "Expectation Failed"; break;
      case 500: httpHeader += "Internal Server Error"; break;
      case 501: httpHeader += "Not Implemented"; break;
      case 502: httpHeader += "Bad Gateway"; break;
      case 503: httpHeader += "Service Unavailable"; break;
      case 504: httpHeader += "Gateway Timeout"; break;
      case 505: httpHeader += "HTTP Version Not Supported"; break;
    }

    // Die Lnge des Bodys anhngen
    httpHeader += "\r\n";
    if(bodyLength > 0)
      httpHeader += "Content-length: " + bodyLength + "\r\n";

    // Auerdem ein Header, das den Browser anweist, die Verbindung zu
    // schlieen
    httpHeader += "Connection: close\r\n";

    // Nur wenn alle Daten bekannt sind das Header abschlieen
    if(bodyLength > 0)
      httpHeader += "\r\n";

    return httpHeader;
  }

  // Anfrage an CGI bergeben
  // Mehr zum CGI Standard unter
  // www.w3.org/CGI/
  void useCGI(String executable, String scriptFile, String queryString, HashMap headers, byte[] postData, boolean headerOnly) throws IOException
  {
    // CGIs bentigen eine bestimmte Betriebsumgebung
    // Diese werden als Umgebungsvariablen bergeben
    HashMap myEnvironment = new HashMap(5);
    myEnvironment.put("REDIRECT_STATUS", "200");
    if(headers.containsKey("content-type"))
      myEnvironment.put("CONTENT_TYPE", headers.get("content-type"));
    if(headers.containsKey("content-length"))
      myEnvironment.put("CONTENT_LENGTH", headers.get("content-length"));
    myEnvironment.put("SERVER_SOFTWARE", "JavaHTTPd");
    myEnvironment.put("SERVER_NAME", headers.get("host"));
    myEnvironment.put("SERVER_PORT", String.valueOf(JavaHTTPd.setting_serverPort));
    myEnvironment.put("SERVER_ADDR", this.requestSocket.getLocalAddress().getHostAddress());
    myEnvironment.put("REMOTE_ADDR", this.requestSocket.getInetAddress().getHostAddress());
    myEnvironment.put("DOCUMENT_ROOT", JavaHTTPd.setting_RootDirectory + "htdocs/");
    myEnvironment.put("SCRIPT_FILENAME", JavaHTTPd.setting_RootDirectory + "htdocs/" + scriptFile);
    myEnvironment.put("GATEWAY_INTERFACE", "CGI/1.1");
    myEnvironment.put("SERVER_PROTOCOL", "HTTP/1.1");
    myEnvironment.put("REQUEST_METHOD", postData.length > 0 ? "POST" : "GET");
    myEnvironment.put("QUERY_STRING", queryString);
    myEnvironment.put("REQUEST_URI", scriptFile + (queryString.length() > 0 ? "?" : "") + queryString);
    myEnvironment.put("SCRIPT_NAME", scriptFile);
    myEnvironment.put("PATH_INFO", scriptFile);
    myEnvironment.put("PATH_TRANSLATED", JavaHTTPd.setting_RootDirectory + "/htdocs/" + scriptFile);

    // Die HTTP Header dazuschreiben
    Object RequestHeaderKeys[] = headers.keySet().toArray();
    for(int iHeaderKeys=0; iHeaderKeys<RequestHeaderKeys.length; iHeaderKeys++)
        myEnvironment.put("HTTP_" + ((String)RequestHeaderKeys[iHeaderKeys]).toUpperCase().replaceAll("-", "_"), headers.get(RequestHeaderKeys[iHeaderKeys]));

    // Die HashMap in ein Array konvertieren
    Object headerKeys[] = myEnvironment.keySet().toArray();
    String arrayEnvironment[] = new String[headerKeys.length];
    for(int iHeaderKeys=0; iHeaderKeys<headerKeys.length; iHeaderKeys++)
      arrayEnvironment[iHeaderKeys] = (String)headerKeys[iHeaderKeys] + "=" + (String)myEnvironment.get(headerKeys[iHeaderKeys]);

    // Nachdem das geschehen ist, den betreffenden CGI Prozess ffnen
    Process cgiProcess = Runtime.getRuntime().exec(executable + " " + JavaHTTPd.setting_RootDirectory + "htdocs/" + scriptFile, arrayEnvironment);

    // Und die Postdaten an ihn senden
    cgiProcess.getOutputStream().write(postData);
    cgiProcess.getOutputStream().flush();
    cgiProcess.getOutputStream().close();

    // Nun in aller Seelenruhe (^^) die Daten auslesen & ausgeben
    InputStream cgiInput = cgiProcess.getInputStream();

    // Dem CGI Standard entsprechend mssen erst Header gesendet werden
    String headerData = "";
    long timeOut = Calendar.getInstance().getTime().getTime() + JavaHTTPd.setting_ConnectionTimeout * 1000;
    do
    {
      if(cgiInput.available() > 0)
        headerData += (char)cgiInput.read();

      // Auf End of File prfen
      if(cgiInput.available() < 0)
      {
        this.errorDie(500);
        return;
      }

      // Auf Timeout prfen
      if (Calendar.getInstance().getTime().getTime() > timeOut)
      {
        // 500 = Serverfehler
        this.errorDie(500);
        cgiProcess.destroy();
        return;
      }

      // Die Kontrolle abgeben, um Prozessorleistung zu sparen
      Thread.yield();
    }
    while(headerData.indexOf("\r\n\r\n") == -1);

    // Die HTTP Statuszeile ausgeben
    if(headerData.toLowerCase().indexOf("status: ") != -1)
    {
      int statusCode = Integer.parseInt(headerData.substring(headerData.toLowerCase().indexOf("status: ") + 8, headerData.toLowerCase().indexOf("status: ") + 11));
      HTTPSocketWrite.write(this.genHTTPHeader(statusCode, -1).getBytes());
    }
    else
      HTTPSocketWrite.write(this.genHTTPHeader(200, -1).getBytes());

    // Den Header ausgeben
    HTTPSocketWrite.write(headerData.getBytes());

    // Wenn nur der Header gesendet werden soll hier abbrechen
    if(headerOnly)
    {
      cgiInput.close();
      cgiProcess.destroy();
      HTTPSocketWrite.flush();
      HTTPSocketWrite.close();
      return;
    }

    // Jetzt bis zum EOF lesen & ausgeben
    while(true)
    {
      byte buffer[] = new byte[1024];
      int currentCharCount = cgiInput.read(buffer);
      if(currentCharCount == -1) break; // End of file
      HTTPSocketWrite.write(buffer, 0, currentCharCount);
    }

    // Socket & CGI Programm Schlieen
    HTTPSocketWrite.flush();
    HTTPSocketWrite.close();
    cgiInput.close();
    cgiProcess.destroy();
    return;
  }
}
